
import collections
import copy
import json
import logging

from paho.mqtt.properties import Properties
from paho.mqtt.packettypes import PacketTypes
import paho.mqtt.client 

from sarracenia.moth import Moth
from sarracenia.flowcb.gather import msg_validate,msg_dumps
import ssl
import threading
import time


logger = logging.getLogger(__name__)

default_options = {
   'auto_ack': True,
   'batch' : 25,
   'clean_session': False,
   'no': 0,
   'prefetch': 25,
   'qos' : 1,
   'topicPrefix' : [ 'v03' ]
}


class MQTT(Moth):
    """
       Message Queue Telemetry Transport support.
           talks to an MQTT broker.  Tested with mosquitto. requires MQTTv5

       Concept mapping from AMQP:

           Only supports v03 messages since mqtt3.1 has no headers.


           AMQP -> MQTT topic hierarcy mapping: 
           exchange: xpublic,  topic prefix:   v03 subtopic:   /my/favourite/directory
           results in :   xpublic/v03/my/favourite/directory

           AMQP queue -> MQTT group-id (client-id should map to  queue+instance, I think.)

       deviation from standard: group-id lengths are not checked, neither are client-id's.
           Sarracenia routinely generates such ids with 40 to 50 characters in them.
           the MQTTv3.1 standard specifies a maximum length of 23 characters, as the minimum
           a compliant broker must support.  
           https://www.eclipse.org/lists/mosquitto-dev/msg00433.html#:~:text=It%20is%2065535%20bytes.,the%20limit%20in%20mqtt%20v3.
           Both mosquitto and EMQ supports 65535 chars for that field even in v3.1, so enforcing limit seems counter-productive.

       problems with MQTT:
           lack of explicit acknowledgements (in paho library) means ack's happen without the
           application being able to process the messages... potential for message loss of
           whatever is queued... this is not good for reliability.  
           It looks like just a choice in the paho library python version... other language bindings apparently have it.
           added to bug report: https://github.com/eclipse/paho.mqtt.python/pull/554
           made a pull request: https://github.com/eclipse/paho.mqtt.python/pull/554 
           Implemented support for the modification here. Seems to work fine (not thoroughly tested yet.)

      Why v5 only:

           No shared subscriptions prior to v5. Used extensively. had originally written library with v3
           support just restricting instance=1, but then I read about the re-transmission strategy:

           In MQTT < 5, in QoS==1, one is supposed to resend messages until you get an ack. some implementations
           do it every 20 seconds.  So in a case where you have, say 5 minutes of queueing, each message will be 
           sent 300/20 -> 15 times...  an unhelpful packet storm.


      There is additionally a vulnerability/inefficiency resulting from async message reception:
           when a new message arrive the loop thread started by the paho library will append
           it to the received_messages data structure (protected by a mutex.)
           if the application crashes, the received_messages have not been ack'd to the sender... so it is
           likely that you will just get them later... could use for a shelf for received_messages,
           and sync it once in while... probably not worth it.
 
    """
    def __init__(self, broker, options, is_subscriber):
        """

          MQTT connection instances...
          in AMQP: topic separator is dot (.) in MQTT, the topic separator is slash (/)
          have exchange arguments use protocol specific separator or mapped one?

        """

        super().__init__(broker, options, is_subscriber)

        self.o.update(default_options)
        self.o.update(options)

        logger.setLevel(getattr(logging, self.o['logLevel'].upper()))

        self.proto_version=paho.mqtt.client.MQTTv5

        ebo=1
        if is_subscriber:
            # https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Receive_Maximum
            if not 'receiveMaximum' in self.o:
                 self.o['receiveMaximum'] = int( min( self.o['batch'] + self.o['prefetch'], 65535 ))
            self.__getSetup(self.o) 
        else:
            self.__putSetup(self.o)

        logger.info("note: mqtt support is newish, not very well tested")
        

    def __sub_on_disconnect(client, userdata, rc):
        logger.info( paho.mqtt.client.connack_string(rc) )
        if hasattr(userdata,'pending_messages'):
            lost=len(userdata.pending_messages)
            if lost > 0:
                logger.error( 'message loss! cannot confirm %d messages were published: mids=%s' % (lost, userdata.pending_messages )  )
            else:
                logger.info('clean. no published messages lost.')

    def __mgt_on_connect(client, userdata, flags, rc, properties=None):
        logger.info( "management connection succeeded: rc=%s, flags=%s" % ( paho.mqtt.client.connack_string(rc), flags ) )

    def __sub_on_connect(client, userdata, flags, rc, properties=None):
        logger.info( "rc=%s, flags=%s" % ( paho.mqtt.client.connack_string(rc), flags ) )

        if flags['session present'] != 1:
            logger.error( 'failed to find existing session, no recovery of inflight messages from previous connection' )

        if rc != paho.mqtt.client.MQTT_ERR_SUCCESS:
            return

        # FIXME: enhancement could subscribe accepts multiple (subj, qos) tuples so, could do this in one RTT.
        for binding_tuple in userdata.o['bindings']:
            exchange, prefix, subtopic = binding_tuple
            logger.info( "tuple: %s %s %s" % ( exchange, prefix, subtopic ) )
            subj = '/'.join( ['$share', userdata.o['queue_name'], exchange] + prefix + subtopic )

            (res, mid) = client.subscribe( subj , qos=userdata.o['qos'] )
            logger.info( "subscribed to: %s, mid=%d qos=%s result: %s" % (subj, mid, \
                userdata.o['qos'], paho.mqtt.client.error_string(res)) )

        client.subscribe_in_progress=True

    def __sub_on_subscribe(client, userdata, mid, granted_qos, properties=None):
        logger.info( "subscribe completed mid={} granted_qos={}".format( mid, list( map( lambda x: x.getName(), granted_qos ))) )

    def __pub_on_disconnect(client, userdata, rc ):
        logger.info( paho.mqtt.client.connack_string(rc) )

    def __pub_on_connect(client, userdata, flags, rc, properties=None):
        logger.info( paho.mqtt.client.connack_string(rc) )

    def __pub_on_publish(client, userdata, mid):
        userdata.pending_messages_mutex.acquire()

        if mid in userdata.pending_messages:
            logger.info( 'publish complete. mid={}'.format(mid) )
            userdata.pending_messages.remove(mid)
        else:
            logger.error('BUG: ack for message we do not know we published. mid={}'.format(mid) )

        userdata.pending_messages_mutex.release()

    def __sslClientSetup(self):
        """
          Initializse self.client SSL context, must be called after self.client is instantiated.
          return port number for connection.
      
        """
        if self.broker.scheme[-1] == 's' :
            port=8883
            logger.info('tls_rigour: %s' % self.o['tls_rigour'] )
            self.o['tls_rigour'] = self.o['tls_rigour'].lower()
            if self.o['tls_rigour'] == 'lax':
                    self.tlsctx = ssl.create_default_context()
                    self.tlsctx.check_hostname = False
                    self.tlsctx.verify_mode = ssl.CERT_NONE
     
            elif self.o['tls_rigour'] == 'strict':
                    self.tlsctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
                    self.tlsctx.options |= ssl.OP_NO_TLSv1
                    self.tlsctx.options |= ssl.OP_NO_TLSv1_1
                    self.tlsctx.check_hostname = True
                    self.tlsctx.verify_mode = ssl.CERT_REQUIRED
                    self.tlsctx.load_default_certs()
                    # TODO Find a way to reintroduce certificate revocation (CRL) in the future
                    #  self.tlsctx.verify_flags = ssl.VERIFY_CRL_CHECK_CHAIN
                    #  https://github.com/MetPX/sarracenia/issues/330
            elif self.o['tls_rigour'] == 'normal':
                self.tlsctx = ssl.create_default_context()
            else:
                self.logger.warning( "option tls_rigour must be one of:  lax, normal, strict")
            self.client.tls_set_context(self.tlsctx)
        else:
            port=1883

        if self.broker.port:
           port =  self.broker.port 
        return port

    def __clientSetup(self, options, cid):

        client = paho.mqtt.client.Client( userdata=self, \
            client_id=cid, protocol=paho.mqtt.client.MQTTv5 )

        #self.client.o = options
        client.connected=False
        client.on_connect = MQTT.__sub_on_connect
        client.on_disconnect = MQTT.__sub_on_disconnect
        client.on_message = MQTT.__sub_on_message
        client.on_subscribe = MQTT.__sub_on_subscribe
        client.subscribe_in_progress=True
        # defaults to 20... kind of a mix of "batch" and prefetch... 
        client.max_inflight_messages_set(options['batch']+options['prefetch'])
        client.username_pw_set( self.broker.username, self.broker.password )
        return client

    def __getSetup(self, options):
        """
           Establish a connection to consume messages with.  
        """
        ebo = 1
        while True:
            try: 
                self.new_message_mutex = threading.Lock()

                cs=options['clean_session']
                if ('queue_name' in options ) and ( 'no' in options ):
                     cid = options['queue_name'] + '%02d' % options['no']
                else:
                     #cid = ''.join(random.choice(string.ascii_lowercase) for i in range(10))
                     cid = None
                     cs=True

                props=Properties(PacketTypes.CONNECT)
                props.SessionExpiryInterval=int(self.o['expire'])
                props.ReceiveMaximum=int(self.o['receiveMaximum'])

                if ( not ('no' in options) or options['no'] == 0 ) and 'instances' in options: 
                    logger.info('declare sessions for instances')
                    for i in range(1,options['instances']+1):
                        icid= options['queue_name'] + '%02d' % i 
                        decl_client = self.__clientSetup( options, icid )
                        decl_client.connect( self.broker.hostname, port=self.__sslClientSetup(), \
                           clean_start=False, properties=props )
                        while not decl_client.is_connected():
                            decl_client.loop(1)
                        decl_client.disconnect()
                        decl_client.loop_stop()
                        logger.info('instance declaration for %s done' % icid )

                    self.client = self.__clientSetup( options, cid )
                    self.client.on_connect = MQTT.__mgt_on_connect
                else:
                    self.client = self.__clientSetup( options, cid )

                self.new_message_mutex.acquire()
                self.client.received_messages=[]
                self.new_message_mutex.release()

                if hasattr( self.client, 'auto_ack' ): # FIXME breaking this...
                    self.client.auto_ack( False )
                    logger.debug("Switching off auto_ack for higher reliability. Using explicit acknowledgements." )
                    self.auto_ack=False
                else:
                    logger.warning("paho library without auto_ack support. Loses data every crash or restart." )
                    self.auto_ack=True

                self.client.connect_async( self.broker.hostname, port=self.__sslClientSetup(), \
                       clean_start=False, properties=props )
                self.client.loop_start()
                return

                
            except Exception as err:
                logger.error("failed to {} with {}".format( self.broker.hostname, err))
                logger.error('Exception details: ', exc_info=True)

            if ebo < 60 : ebo *= 2
            time.sleep(ebo)


    def __putSetup(self, options):
        """
           establish a connection to allow publishing. 
        """
        ebo=1
        while True:
            try:

                props=Properties(PacketTypes.CONNECT)
                if self.o['message_ttl'] > 0:
                    props.MessageExpiryInterval=int(self.o['message_ttl'])

                self.pending_messages_mutex= threading.Lock()
                self.pending_messages=collections.deque()

                self.client = paho.mqtt.client.Client( protocol=self.proto_version, userdata=self) 
                self.client.on_connect = MQTT.__pub_on_connect
                self.client.on_disconnect = MQTT.__pub_on_disconnect
                self.client.on_publish = MQTT.__pub_on_publish
                #dunno if this is a good idea.
                self.client.max_queued_messages_set(options['prefetch'])
                self.client.username_pw_set( self.broker.username, self.broker.password )
                res = self.client.connect( options['broker'].hostname, port=self.__sslClientSetup(), properties=props  )
                logger.info( 'connecting to %s, res=%s' % (options['broker'].hostname, res ) )

                self.client.loop_start()
                return

            except Exception as err:
                logger.error("failed to {} with {}".format( self.broker.hostname, err))
                logger.error('Exception details: ', exc_info=True)

            if ebo < 60 : ebo *= 2
            time.sleep(ebo)
               
        
    def __sub_on_message(client, userdata, msg):
        """
          callback to append messages received to new queue.
          MQTT only supports v03 messages, so always assumed to be JSON encoded.

          FIXME: locking here is expensive... would like to group them ... 1st draft.
             could do a rotating set of batch size lists, and that way just change counters.
             and not lock individual messages  list[1], list[2], ... consumer consumes old lists...
             no locking needed, except to increment list index.... later.
        """
       
        #dm = userdata._msgDecode(msg)
        #if dm:
        logger.info( "Message received: %d, %s %s" % (msg.mid, msg.topic, msg.payload.decode('utf-8')) )
        userdata.new_message_mutex.acquire()
        client.received_messages.append( msg )
        userdata.new_message_mutex.release()

    def putCleanUp(self):
        self.client.disconnect()
        self.client.loop_stop()
        pass

    def getCleanUp(self):         

        if ( not ('no' in self.o) or self.o['no'] == 0 ) and 'instances' in self.o: 
            props=Properties(PacketTypes.CONNECT)
            props.SessionExpiryInterval=1
            logger.info('cleanup sessions for instances')
            for i in range(1,self.o['instances']+1):
                icid= self.o['queue_name'] + '%02d' % i 
                myclient = self.__clientSetup( options, icid )
                myclient.connect( self.broker.hostname, port=self.__sslClientSetup(), \
                   myclean_start=True, properties=props )
                while not self.client.is_connected():
                    myclient.loop(0.1)
                myclient.disconnect()
                logger.info('instance deletion for %02d done' % i )

        self.client.disconnect()
        self.client.loop_stop()
        pass

    def _msgDecode(self, mqttMessage ):
        """
           decode MQTT message (protocol specific thingamabob) into sr3 one (python dictionary)
        """
        try:
            message = json.loads(mqttMessage.payload.decode('utf-8'))
        except Exception as ex:
            logger.error( "ignored malformed message: %s" % mqttMessage.payload.decode('utf-8') )
            logger.error("decode error" % err)
            logger.error('Exception details: ', exc_info=True)
            return None

        subtopic=mqttMessage.topic.split('/')
         
        if subtopic[0] != self.o['topicPrefix'][0]:
            message['exchange'] = subtopic[0]
            message['subtopic'] = subtopic[1+len(self.o['topicPrefix']):]
        else:
            message['subtopic'] = subtopic[len(self.o['topicPrefix']):]
        message['ack_id'] = mqttMessage.mid
        message['local_offset'] = 0
        message['_deleteOnPost'] = set( [ 'exchange', 'local_offset', 'ack_id', 'subtopic' ] )

        if msg_validate( message ):
           return message
        else:
           return None

    def newMessages(self):
        """
           return new messages.

           FIXME: hate the locking... too fine grained, especially in on_message... just a 1st shot.

        """
        self.new_message_mutex.acquire()

        if len(self.client.received_messages) > self.o['batch'] :
            mqttml=self.client.received_messages[0:self.o['batch']]
            self.client.received_messages=self.client.received_messages[self.o['batch']:]
        else:
            mqttml=self.client.received_messages
            self.client.received_messages=[]

        self.new_message_mutex.release()

        ml = list(filter( None, map( self._msgDecode, mqttml ) ))
        return ml

    def getNewMessage(self):

        self.new_message_mutex.acquire()

        if len(self.client.received_messages) > 0: 
            m=self.client.received_messages[0]
            self.client.received_messages=self.client.received_messages[1:]
        else:
            m=None
        self.new_message_mutex.release()

        return self._msgDecode(m)
    
    def ack(self, m):
        if (not self.auto_ack) and ('ack_id' in m):
            logger.info('mid=%d' % m['ack_id'] )
            self.client.ack(m['ack_id'])
            del m['ack_id']
            m['_deleteOnPost'].remove('ack_id')

    def putNewMessage(self, body, content_type='application/json', exchange=None):
        """
          publish a message.
        """
        if self.is_subscriber:  #build_consumer
           logger.error("publishing from a consumer")
           return False
 
        if not self.client.is_connected():
           logger.error("no connection to publish to broker with")
           return False

        #body = copy.deepcopy(bd)

        if '_deleteOnPost' in body:
            # FIXME: need to delete because building entire JSON object at once.
            # makes this routine alter the message. Ideally, would use incremental
            # method to build json and _deleteOnPost would be a guide of what to skip.
            # library for that is jsonfile, but not present in repos so far.
            for k in body['_deleteOnPost']:
                if k == 'subtopic' : continue
                if k in body:
                    del body[k]
            del body['_deleteOnPost']
  
         
        if not exchange:
             if (type(self.o['exchange']) is list):
                  if (len(self.o['exchange']) > 1):
                      if 'post_exchange_split' in self.o:
                          # FIXME: assert ( len(self.o['exchange']) == self.o['post_exchange_split'] )
                          #        if that isn't true... then there is something wrong... should we check ?
                          idx = sum(
                              bytearray(body['integrity']['value'],
                                        'ascii')) % len(self.o['exchange'])
                          exchange = self.o['exchange'][idx]
                      else:
                          logger.error(
                              'do not know which exchange to publish to: %s' %
                              self.o['exchange'])
                          return
                  else:
                      exchange = self.o['exchange'][0]
             else:
                  exchange = self.o['exchange']
  
        # FIXME: might 
        topic= '/'.join( [ exchange ] + self.o['topicPrefix'] + body['subtopic']  )
        topic = topic.replace('#', '%23')

        del body['subtopic']
        props=Properties(PacketTypes.PUBLISH)
        # https://docs.oasis-open.org/mqtt/mqtt/v5.0/os/mqtt-v5.0-os.html#_Toc3901111
        props.PayloadFormatIndicator = 1 # designates UTF-8
        props.ContentType = 'application/json'

        while True:
            try:
                info = self.client.publish( topic=topic, payload=json.dumps(body), qos=1, properties=props) 
                if info.rc == paho.mqtt.client.MQTT_ERR_SUCCESS: 
                    self.pending_messages_mutex.acquire()
                    self.pending_messages.append(info.mid)
                    self.pending_messages_mutex.release()

                    logger.info("published mid={} {} to under: {} ".format(info.mid, body, topic))
                    return True #success...

            except Exception as ex:
                logger.error('Exception details: ', exc_info=True)

  
  
    def close(self):
        logger.info('closing')
        if self.client.is_connected():
            if self.is_subscriber and self.client.subscribe_in_progress:
                time.sleep(0.1) 

            if hasattr(self,'pending_messages'):
                while len(self.pending_messages) > 0:
                    logger.info('waiting for last messages to publish')
                    time.sleep(0.1)    
                logger.info('no more pending messages')
            self.client.disconnect()
